from manga_py.provider import Provider
from .helpers.std import Std


class InMangaCom(Provider, Std):
    __local_storage = None

    def get_archive_name(self) -> str:
        return self.normal_arc_name(self.get_chapter_index())

    def get_chapter_index(self) -> str:
        return str(self.chapter['Number'])

    def get_main_content(self):
        if not self.__local_storage.get('uri_hex', False):
            self.get_manga_name()
        url = '{}/chapter/getall?mangaIdentification={}'.format(
            self.domain,
            self.__local_storage['uri_hex']
        )
        data = self.json.loads(self.http_get(url))['data']
        return self.json.loads(data)

    def get_manga_name(self) -> str:
        url = self.get_url()
        test = self.re.search(r'/ver/manga/[^/]+/\d+/[^/]+', url)
        if test:
            content = self._elements('.chapterControlsContainer label.blue a.blue')[0]
            url = self.domain + content.get('href')
        manga_name, uri_hex = self.re.search('/ver/manga/([^/]+)/([^/]+)', url).groups()
        self.__local_storage['manga_name'] = manga_name
        self.__local_storage['uri_hex'] = uri_hex
        return self.__local_storage['manga_name']

    @staticmethod
    def __sort_chapters(items, reverse=False):
        return sorted(items, key=lambda i: float(i['FriendlyChapterNumber']), reverse=reverse)

    def get_chapters(self):
        items = self.content['result']
        return self.__sort_chapters(items, True)

    def prepare_cookies(self):
        self.__local_storage = {}

    def _make_url(self, chapter):
        return '{}/ver/manga/{}/{}/{}'.format(
            self.domain,
            self.manga_name,
            chapter['FriendlyChapterNumber'],
            chapter['Identification']
        )

    def get_files(self):
        files_url = '{}/page/getPageImage/?identification={}'
        url = self._make_url(self.chapter)
        images = self.html_fromstring(url, '.PagesContainer img.ImageContainer')

        domain = self.domain
        return [files_url.format(domain, i.get('id')) for i in images]

    def get_cover(self):
        idx = self.__local_storage['uri_hex']
        return '{}/manga/getMangaImage?identification={}'.format(self.domain, idx)

    def book_meta(self) -> dict:
        # todo meta
        pass

    def chapter_for_json(self):
        return self._make_url(self.chapter)


main = InMangaCom



from manga_py.provider import Provider
from .helpers.std import Std


class WMangaRu(Provider, Std):

    def get_archive_name(self) -> str:
        idx = self.get_chapter_index().split('-')
        return self.normal_arc_name(idx)

    def get_chapter_index(self) -> str:
        selector = r'/manga_chapter/[^/]+/(\d+)/(\d+)'
        idx = self.re.search(selector, self.chapter).groups()
        return '{}-{}'.format(*idx)

    def get_main_content(self):
        return self._get_content('{}/starter/manga_byid/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/starter/manga_[^/]+/([^/]+)')

    def get_chapters(self):
        return self._elements('td div div div td > a')[::-1]

    def get_files(self):
        parser = self.html_fromstring(self.chapter)
        return self._images_helper(parser, 'td a.gallery', 'href')

    def get_cover(self):
        pass  # FIXME HOME

    def book_meta(self) -> dict:
        # todo meta
        pass


main = WMangaRu



from manga_py.provider import Provider
from .helpers.std import Std


class AsmHentaiCom(Provider, Std):
    _content = None

    def get_archive_name(self) -> str:
        return 'archive'

    def get_chapter_index(self) -> str:
        return '0'

    def get_main_content(self):
        if self._content is None:
            self._content = self.http_get(self.get_url())
        return self._content

    def get_manga_name(self) -> str:
        title = self.text_content(self.content, '.info > h1,title')
        if ~title.find(' Page '):
            title = self.re.search(r'(.+) Page ', title).group(1)
        return title

    def get_chapters(self):
        url = self.get_url()
        if ~url.find('/g/'):
            url = self._elements('.gallery > div > a')[0].get('href')
        return [url]

    def get_files(self):
        content = self.http_get(self.chapter)
        src = self.re.search(r'\$\(\[[\'"]//(.+)/[\'"]', content).group(1)
        pages = self.re.search(r'var +Pages ?=.*?(\d+)', content).group(1)
        result = []
        http = self.re.search('(https?):', self.get_url()).group(1)
        for i in range(int(pages)):
            result.append('{}://{}/{}.jpg'.format(http, src, 1 + i))
        return result

    def get_cover(self) -> str:
        return self._cover_from_content('.cover > a > img')

    def book_meta(self) -> dict:
        pass


main = AsmHentaiCom



from urllib.parse import unquote, quote

from manga_py.provider import Provider
from .helpers.std import Std


class MangaRussiaCom(Provider, Std):

    @staticmethod
    def path_url(url):
        return quote(unquote(url)).replace('%3A//', '://')

    def get_archive_name(self) -> str:
        idx = self.get_chapter_index().split('-')
        return self.normal_arc_name(idx)

    def get_chapter_index(self) -> str:
        chapter = self.chapter
        result = self.re.search(r'\+(\d+)\+-\+(\d+)', chapter).groups()
        return '-'.join(*result)

    def get_main_content(self):
        url = '{}/manga/{}.html'.format(self.domain, quote(self.manga_name))
        self._storage['referer'] = self.path_url(self.get_url())
        return self.http_get(url)

    def __name(self, url):
        return self.re.search(r'/manga/(.+)\.html', url).group(1)

    def get_manga_name(self) -> str:
        url = self.get_url()
        if self.re.search('/manga/', url):
            name = self.__name(url)
        else:
            url = self.html_fromstring(url, '.sitemaplist .red', 0).get('href')
            name = self.__name(url)
        return unquote(name)

    def get_chapters(self):
        return self._elements('.chapterlist .col1 a')

    def _get_img(self, parser):
        img = parser.cssselect('img#comicpic')[0]
        urls = [img.get('src')]
        onload = img.get('onload')
        if ~onload and onload.find('(\''):
            urls.append(self.re.search('\(\'(.+)\'\)', onload).group(1))
        return urls

    def get_files(self):
        parser = self.html_fromstring(self.chapter)
        result = parser.cssselect('select#page option + option')
        images = self._get_img(parser)
        for n, i in enumerate(result):
            if n and n % 2:
                parser = self.html_fromstring(i.get('value'))
                images += self._get_img(parser)
        return images

    def get_cover(self):
        self._cover_from_content('.bookfrontpage > a > img')

    def before_download_chapter(self):
        self._storage['referer'] = self.path_url(self.chapter)

    def book_meta(self) -> dict:
        # todo meta
        pass


main = MangaRussiaCom



from manga_py.crypt import AcQqComCrypt
from manga_py.provider import Provider
from .helpers.std import Std


class AcQqCom(Provider, Std):
    _decoder = None
    _re = None

    def get_archive_name(self) -> str:
        return self.normal_arc_name(self.get_chapter_index())

    def get_chapter_index(self) -> str:
        return self.re.search(r'/cid/(\d+)', self.chapter).group(1)

    def get_main_content(self):
        content = self._storage.get('main_content', None)
        if content is not None:
            return content
        idx = self._get_name(r'/id/(\d+)')
        return self.http_get('{}/Comic/comicInfo/id/{}'.format(self.domain, idx))

    def get_manga_name(self) -> str:
        return self.text_content(self.content, '.works-intro-title strong', 0)

    def get_chapters(self):
        return self._elements('.chapter-page-all li a')[::-1]

    def get_files(self):
        content = self.http_get(self.chapter)
        data = self._re.search(content).group(1)
        data = self._decoder.decode(data)
        return [i.get('url') for i in data.get('picture', [])][0:1]

    def get_cover(self) -> str:
        return self._cover_from_content('.works-cover img')

    def prepare_cookies(self):
        self._re = self.re.compile(r'var\s+DATA\s*=\s*[\'"](.*?)[\'"]')
        self._decoder = AcQqComCrypt(self)
        self._base_cookies()

    def book_meta(self) -> dict:
        result = {
            'author': self.text_content(self.content, '.works-intro-digi em'),
            'rating': self.text_content(self.content, 'p.ui-left strong'),
            'cover': self.get_cover(),
            'annotation': self.text_content(self.content, '.works-intro-short'),
            'language': 'cn',
        }

        return result


main = AcQqCom



from manga_py.provider import Provider
from .helpers.std import Std


class HGameCGCom(Provider, Std):
    __img_selector = '#thumbnails > div .col-thumbnail'
    __img_count = 0

    def get_archive_name(self) -> str:
        return 'page_{:0>3}'.format(self.chapter)

    def get_chapter_index(self) -> str:
        return '0'

    def get_main_content(self):
        return self._get_content('{}/index/category/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/index/category/([^/]+)')

    def get_chapters(self):
        images = self.document_fromstring(self.content, self.__img_selector)
        self.__img_count = len(images)
        pages = self.document_fromstring(self.content, '.pagination li > a')
        pages_count = 0
        if pages:
            pages_count = self.re.search('/start-(\d+)', pages[-1].get('href')).group(1)
            pages_count = int(int(pages_count) / self.__img_count)
        return range(1, pages_count + 2)[::-1]

    def __tmb_to_img(self, tmbs):
        url = '{}/action.php?id={}&part=e'
        imgs = []
        re = self.re.compile(r'/picture/(\d+)')
        for i in tmbs:
            idx = re.search(i.get('href')).group(1)
            imgs.append(url.format(self.domain, idx))
        return imgs

    def get_files(self):
        pages_url = self.get_url() + '/start-{}'
        if self.chapter > 1:
            offset = self.__img_count * (self.chapter - 1)
            images = self.html_fromstring(pages_url.format(offset), self.__img_selector)
        else:
            images = self.document_fromstring(self.content, self.__img_selector)
        return self.__tmb_to_img(images)

    def get_cover(self) -> str:
        # return self._cover_from_content('.cover img')
        pass

    def book_meta(self) -> dict:
        pass

    def chapter_for_json(self):
        return self.get_url()


main = HGameCGCom



from manga_py.provider import Provider
from .helpers.std import Std
from manga_py.crypt import mangago_me
from manga_py.fs import rename, unlink


class MangaGoMe(Provider, Std):
    _enc_images = None
    _crypt = None

    def get_archive_name(self) -> str:
        idx = self.get_chapter_index().split('-')
        tp = self.re.search('/(\w{1,4})/[^/]*?\d+', self.chapter)
        idx = [self.chapter_id, idx[-1]]  # do not touch this!
        if tp:
            idx.append(tp.group(1))
        return self.normal_arc_name(idx)

    def get_chapter_index(self) -> str:
        selector = r'/\w{1,4}/[^/]*?(\d+)(?:[^\d]+(\d+))?'
        idx = self.re.search(selector, self.chapter).groups()
        if idx[1] is not None:
            fmt = '{}-{}'
            return fmt.format(*idx)
        return idx[0]

    def get_main_content(self):
        return self._get_content(self.get_url())

    def get_manga_name(self) -> str:
        return self._get_name(r'/read-manga/([^/]+)/')

    def get_chapters(self):
        content = self._elements('#information')
        if not content:
            return []
        chapters = content[0].cssselect('#chapter_table a.chico')
        raws = content[0].cssselect('#raws_table a.chicor')
        return chapters + raws

    def prepare_cookies(self):
        self._crypt = mangago_me.MangaGoMe()
        self.cf_protect(self.domain)

    def get_files(self):
        self._enc_images = {}
        content = self.http(True, {
            'referer': self.chapter,
            'cookies': self.http().cookies,
            'user_agent': self.http().user_agent
        }).get(self.chapter)
        re = self.re.search(r"imgsrcs\s*=\s*['\"](.+)['\"]", content)
        if not re:
            return []
        items = self._crypt.decrypt(re.group(1))
        return items.split(',')

    def before_file_save(self, url, idx):
        if ~url.find('/cspiclink/'):
            self._enc_images[idx] = url
        return url

    def after_file_save(self, _path: str, idx: int):
        url = self._enc_images.get(idx, None)
        if url is not None:
            _dst = _path[:_path.rfind('.')] + '_' + _path[_path.rfind('.'):]
            self._crypt.puzzle(_path, _dst, url)
            unlink(_path)
            rename(_dst, _path)

    def get_cover(self):
        return self._cover_from_content('#information .cover img')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = MangaGoMe



from .helpers.std import Std
from .shogakukan_tameshiyo_me import ShogakukanTameshiyoMe


class ShogakukanCoJp(ShogakukanTameshiyoMe, Std):

    def get_archive_name(self) -> str:
        return self.normal_arc_name([
            self.get_chapter_index(),
            self.re.search(r'/(\d+)', self.chapter).group(1)
        ])

    def get_chapter_index(self) -> str:
        return str(self.chapter_id)

    def get_main_content(self):
        types = [
            'magazines/series',
            'books'
        ]
        re = r'(/(?:{})/\d+)'.format('|'.join(types))
        url = self.re.search(re, self.get_url()).group(1)
        return self.http_get(self.domain + url)

    def get_manga_name(self) -> str:
        return self._get_name(r'/(?:series|books)/(\d+)')

    def get_chapters(self):
        return self._elements('a[href*="shogakukan.tameshiyo.me"]')  # todo: watch this

    def get_cover(self) -> str:
        img = self._cover_from_content('.mainimg01')
        if not img:
            img = self._cover_from_content('.image01 > img', 'data-original')
        return img

    def book_meta(self) -> dict:
        # todo meta
        pass


main = ShogakukanCoJp



from manga_py.provider import Provider
from .helpers.std import Std


class MangaNeloCom(Provider, Std):

    def get_archive_name(self) -> str:
        idx = self.get_chapter_index()
        return self.normal_arc_name(idx.split('.', 2))

    def get_chapter_index(self) -> str:
        re = self.re.compile(r'/chapter_(\d+(?:\.\d+)?)')
        return re.search(self.chapter).group(1)

    def get_main_content(self):
        return self._get_content('{}/manga/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/(?:manga|chapter)/([^/]+)')

    def get_chapters(self):
        return self._elements('.chapter-list a')

    def get_files(self):
        parser = self.html_fromstring(self.chapter)
        images = self._images_helper(parser, '#vungdoc img')
        if not len(images):
            images = self._images_helper(parser, '.vung_doc img')
        return images

    def get_cover(self) -> str:
        return self._cover_from_content('.manga-info-pic img')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = MangaNeloCom



from .gomanga_co import GoMangaCo


class JapScanCom(GoMangaCo):
    _name_re = r'\.c(?:om|c)/[^/]+/([^/]+)/'
    _content_str = '{}/mangas/{}/'
    _chapters_selector = '#liste_chapitres ul li a'

    def get_archive_name(self) -> str:
        idx = self.chapter_id, self.get_chapter_index()
        return self.normal_arc_name(idx)

    def get_chapter_index(self) -> str:
        selector = r'\.c(?:om|c)/[^/]+/[^/]+/(\d+)/'
        url = self.chapter
        return self.re.search(selector, url).group(1)

    def get_files(self):
        img_selector = '#image'
        n = self.http().normalize_uri
        parser = self.html_fromstring(self.chapter)
        pages = parser.cssselect('#pages option + option')
        images = self._images_helper(parser, img_selector)
        for i in pages:
            parser = self.html_fromstring(n(i.get('value')))
            images += self._images_helper(parser, img_selector)
        return images

    def get_cover(self) -> str:
        pass


main = JapScanCom



from urllib.parse import unquote_plus

from .hentaifox_com import HentaiFoxCom


class HitomiLa(HentaiFoxCom):
    _idx_re = r'/(?:galleries|reader)/(\d+)'
    _url_str = '{}/galleries/{}.html'
    _name_selector = '.dj-gallery h1 a'
    _cdn = 'http://0a.hitomi.la/galleries/'

    def get_manga_name(self):
        name = super().get_manga_name()
        return unquote_plus(name.split('|')[0].strip())

    def get_files(self):
        idx = self._get_name(self._idx_re)
        url = 'http://ltn.hitomi.la/galleries/{}.js'.format(idx)
        images = self.re.search(r'(\[.+\])', self.http_get(url))
        images = self.json.loads(images.group(1))
        p = '{}{}/'.format(self._cdn, idx)
        return [p + i.get('name') for i in images]


main = HitomiLa



from manga_py.provider import Provider
from .helpers.std import Std


class MangaForAllCom(Provider, Std):

    def get_archive_name(self) -> str:
        idx = self.get_chapter_index().split('-')
        return 'vol_{:0>3}-{}_{}'.format(
            *self._idx_to_x2(idx),
            self.chapter_id
        )

    def get_chapter_index(self) -> str:
        re = self.re.compile(r'-(\d+(?:\.\d+)?)-')
        ch = self.chapter
        return '-'.join(re.search(ch).group(1).split('.'))

    def get_main_content(self):
        return self._get_content('{}/m/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/m/([^/]+)')

    def get_chapters(self):
        return self._elements('.Chapters ul.list-unstyled > li a')

    def get_files(self):
        parser = self.html_fromstring(self.chapter)
        return self._images_helper(parser, '.container ul.list-unstyled > li img')

    def get_cover(self) -> str:
        return self._cover_from_content('meta[property="og:image"]', 'content')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = MangaForAllCom



from .mangago_me import MangaGoMe


class RocacaCom(MangaGoMe):

    def get_archive_name(self) -> str:
        return self.normal_arc_name(self.get_chapter_index().split('-'))

    def get_chapter_index(self) -> str:
        re = self.re.search(r'/.+?chapter-(\d+(?:\.\d+)?)', self.chapter)
        if not re:
            re = self.re.search(r'/.+?Ch(\d+(?:\.\d+)?)', self.chapter)
        if not re:
            re = self.re.search(r'/c(\d+(?:\.\d+)?)', self.chapter)
        if not re:
            re = self.re.search(r'/(\d+(?:\.\d+)?)/', self.chapter)
        return re.group(1).replace('.', '-')

    def get_main_content(self):
        return self._get_content('{}/manga/{}/')

    def get_manga_name(self) -> str:
        return self._get_name('/manga/([^/]+)')

    def get_chapters(self):
        return self._elements('.section-list > table a.chico')

    def get_cover(self) -> str:
        return self._cover_from_content('.cartoon-intro > img.pic')


main = RocacaCom



from .gomanga_co import GoMangaCo


class LectorKirishimaFanSubCom(GoMangaCo):
    _name_re = '/(?:reader/)?(?:series|read)/([^/]+)/'
    _content_str = '{}/lector/series/{}/'


main = LectorKirishimaFanSubCom



from manga_py.provider import Provider
from .helpers.std import Std


class DarkSkyProjectsOrg(Provider, Std):

    def get_archive_name(self) -> str:
        return self.normal_arc_name([
            self.chapter_id,
            self.get_chapter_index()
        ])

    def get_chapter_index(self) -> str:
        ch = self.chapter
        return self.re.search('/biblioteca/[^/]+/([^/]+)', ch).group(1)

    def get_main_content(self):
        return self._get_content('{}/biblioteca/{}')

    def get_manga_name(self) -> str:
        return self._get_name('/biblioteca/([^/]+)')

    def get_chapters(self):
        return self._elements('.chapters h5 a')

    def get_files(self):
        parser = self.html_fromstring(self.chapter)
        return self._images_helper(parser, 'data-src')

    def get_cover(self) -> str:
        return self._cover_from_content('.boxed > .img-responsive')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = DarkSkyProjectsOrg



import re
from os import path
from sys import stderr

from lxml.html import HtmlElement

from manga_py.http import Http
from manga_py.image import Image


class Base:
    _storage = None
    _params = None
    _image_params = None
    _http_kwargs = None
    __http = None

    def __init__(self):

        self._storage = {
            'cookies': {},
            'main_content': None,
            'chapters': [],
            'current_chapter': 0,
            'current_file': 0,
            'proxies': {},
            'domain_uri': None,
        }
        self._params = {
            'destination': 'Manga',
            'cf-protect': False,
        }
        self._image_params = {
            'crop': (0, 0, 0, 0),
            # 'crop': (left, upper, right, lower)
            'auto_crop': False,
            # 'auto_crop': True,
        }
        self._http_kwargs = {}

    def _archive_type(self):
        arc_type = 'zip'
        if self._params['cbz']:
            arc_type = 'cbz'
        return arc_type

    def get_url(self):
        return self._params['url']

    @property
    def domain(self) -> str:
        try:
            if not self._storage.get('domain_uri', None):
                self._storage['domain_uri'] = re.search('(https?://[^/]+)', self._params['url']).group(1)
            return self._storage.get('domain_uri', '')
        except Exception:
            print('url is broken!', file=stderr)
            exit()

    @staticmethod
    def image_auto_crop(src_path, dest_path=None):
        image = Image(src_path=src_path)
        image.crop_auto(dest_path=dest_path)
        image.close()

    def image_manual_crop(self, src_path, dest_path=None):  # sizes: (left, top, right, bottom)
        if isinstance(self._image_params['crop'], tuple) != (0, 0, 0, 0):
            image = Image(src_path=src_path)
            image.crop_manual_with_offsets(offsets=self._image_params['crop'], dest_path=dest_path)
            image.close()

    def _build_http_params(self, params):
        if params is None:
            params = {}
        params.setdefault('allow_webp', not self._params.get('disallow_webp', None))
        params.setdefault('referer', self._storage.get('referer', self.domain))
        params.setdefault('user_agent', self._get_user_agent())
        params.setdefault('proxies', self._storage.get('proxies', None))
        params.setdefault('cookies', self._storage.get('cookies', None))
        params.setdefault('kwargs', self._http_kwargs)
        return params

    def http(self, new=False, params=None) -> Http:
        http_params = self._build_http_params(params)
        if new:
            http = Http(**http_params)
            return http
        elif not self.__http:
            self.__http = Http(**http_params)
        return self.__http

    def http_get(self, url: str, headers: dict = None, cookies: dict = None):
        return self.http().get(url=url, headers=headers, cookies=cookies)

    def http_post(self, url: str, headers: dict = None, cookies: dict = None, data=()):
        return self.http().post(url=url, headers=headers, cookies=cookies, data=data)

    def _get_user_agent(self):
        ua_storage = self._storage.get('user_agent', None)
        ua_params = self._params.get('user_agent', None)
        if self._params.get('cf_protect', False):
            return ua_storage
        return ua_params

    @property
    def chapter_id(self):
        return self._storage.get('current_chapter', 0)

    @classmethod
    def __normalize_chapters(cls, n, element):
        if isinstance(element, HtmlElement):
            return n(element.get('href'))
        if isinstance(element, str):
            return n(element)
        return element

    def _prepare_chapters(self, chapters):
        n = self.http().normalize_uri
        items = []
        if chapters and len(chapters):
            for i in chapters:
                url = self.__normalize_chapters(n, i)
                items.append(url)
        return items

    @property
    def chapter(self):
        return self._storage['chapters'][self._storage['current_chapter']]

    def get_current_file(self):
        return self._storage['files'][self._storage['current_file']]

    def book_meta(self) -> dict:
        return {}

#     def _arc_meta_info(self):
#         info = self.book_meta()
#
#         # {genre}
#         xml = """
# <book-info>
#     {author}
#     {title}
#     {annotation}
#     {keywords}
#     {cover}
#     {rating}
# </book-info>
# """
#
#         key_vars = {
#             'author': 'author',
#             'title': 'book-title',
#             # 'genre': 'genre',
#             'annotation': 'annotation',
#             'keywords': 'keywords',
#             'cover': 'coverpage',
#             'rating': 'content-rating',
#         }
#
#         result = {}
#
#         for i in info:
#             value = info.get(i, None)
#             result[i] = ''
#             if value is not None:
#                 result[i] = '<{0}>{1}</{0}>'.format(key_vars[i], value)
#
#         return xml.format(**result)

    def _image_name(self, idx, filename):
        if idx is None:
            idx = self._storage['current_file']
        fn, extension = path.splitext(filename)
        _path = '{:0>3}_{}'.format(idx, fn)
        if self._params['rename_pages']:
            _path = '{:0>3}'.format(idx)
        return _path + extension

    def chapter_for_json(self) -> str:
        return self.chapter

    def put_info_json(self, meta):
        # manga_name, url, directory
        pass



from PIL import Image


class Puzzle:
    need_copy_orig = False

    __x = 0
    __y = 0
    __multiply = 1
    __matrix = None
    __image_src = None
    __image_dst = None
    __block_width = None
    __block_height = None

    def __init__(self, x: int, y: int, matrix: dict, multiply: int = 1):
        self.__x = x
        self.__y = y
        self.__matrix = matrix
        self.__multiply = multiply

    def de_scramble(self, path_src: str, path_dst: str):
        self.__image_src = Image.open(path_src, 'r')
        self._process()
        self.__image_dst.save(path_dst)
        self.__image_src.close()
        self.__image_dst.close()

    def _process(self):
        self.__image_dst = Image.new(self.__image_src.mode, self.__image_src.size)
        self._calc_block_size()
        self._check_copy_orig_image()
        self._solve_matrix()

    def _check_copy_orig_image(self):
        if self.need_copy_orig:
            self.__image_dst.paste(self.__image_src)

    def _calc_block_size(self):
        if self.__multiply <= 1:
            self.__block_width = int(self.__image_src.size[0] / self.__x)
            self.__block_height = int(self.__image_src.size[1] / self.__y)
        else:
            self.__block_width = self.__multiply * int(self.__image_src.size[0] / self.__y / self.__multiply)
            self.__block_height = self.__multiply * int(self.__image_src.size[1] / self.__x / self.__multiply)

    def _src_rect(self, i):
        row = int(i / self.__x)
        col = i - row * self.__x
        x = col * self.__block_width
        y = row * self.__block_height
        return x, y, x + self.__block_width, y + self.__block_height

    def _dst_rect(self, i):
        row = int(self.__matrix[i] / self.__x)
        col = self.__matrix[i] - row * self.__y
        x = col * self.__block_width
        y = row * self.__block_height
        return x, y, x + self.__block_width, y + self.__block_height

    def _solve_matrix(self):
        for i in range(self.__x * self.__y):
            src_rect = self._src_rect(i)
            dst_rect = self._dst_rect(i)
            region = self.__image_src.crop(src_rect)
            self.__image_dst.paste(region, dst_rect)



from urllib.parse import unquote_plus

from manga_py.provider import Provider
from .helpers.std import Std


class HitMangaEu(Provider, Std):
    _n = None
    postfix = None
    main_domain = 'http://www.mymanga.io'
    api_url = 'http://www.hitmanga.eu/listener/'

    def get_archive_name(self) -> str:
        idx = self.get_chapter_index()
        return self.normal_arc_name(idx)

    def get_chapter_index(self) -> str:
        chapter = self.chapter
        idx = self.re.search('[^/]+/[^/]+/[^/]+?-([^/]+)', chapter)
        return idx.group(1)

    def get_main_content(self):
        return self._content(self._get_content('{}/mangas/{}/'))

    def get_manga_name(self) -> str:
        url = self.get_url()
        re = '{}/([^/]+)'
        if ~url.find('/mangas/'):
            re = '{}/mangas/([^/]+)'
        re = re.format(self.postfix)
        return self.re.search(re, url).group(1)

    def get_chapters(self):
        return self._elements('.listchapseries li a.follow:not(.ddl)')

    def _content(self, url):
        return self.re.sub(r'<!--.+?--!>', '', self.http_get(url))

    def get_files(self):
        chapter = self.chapter
        img = self.document_fromstring(self._content(chapter), '#chpimg', 0).get('src')
        idx = self.get_chapter_index()
        items = self.http_post(url=self.api_url, data={
            'number': unquote_plus(idx),
            'permalink': self.manga_name,
            'type': 'chap-pages',
        })
        if items == '0':
            return []
        items = items.split('|')
        return [self._n(i, img) for i in items]

    def get_cover(self) -> str:
        return self._cover_from_content('#picture img')

    def prepare_cookies(self):
        domain = self.domain.split('.')
        self.postfix = r'\.' + domain[-1]
        n = self.http().normalize_uri
        self._n = lambda u, r: n(u, r)

    def book_meta(self) -> dict:
        # todo meta
        pass


main = HitMangaEu



import tempfile
from os import name as os_name, getpid, makedirs, walk
from pathlib import Path
from shutil import move
from json import loads as json_loads
from shutil import rmtree

__dir_name__ = '.PyMangaDownloader'


def mark_as_hidden(_path: str):
    try:
        from ctypes import windll
        windll.kernel32.SetFileAttributesW(_path, 2)
    except Exception:
        pass


def get_temp_path(*args) -> str:
    temp = 'temp_%s' % getpid()
    return path_join(tempfile.gettempdir(), __dir_name__, temp, *args)


def root_path() -> str:
    #      fs.py/manga_py/../
    file = Path(__file__).resolve()
    return str(file.parent.parent)


def get_util_home_path() -> str:
    if os_name == 'nt':
        home = path_join(str(Path.home()), 'AppData', 'Roaming', __dir_name__)
    else:
        home = path_join(str(Path.home()), __dir_name__)
    make_dirs(home)
    return str(home)


def make_dirs(directory):
    is_dir(directory) or makedirs(directory)


def remove_file_query_params(name, save_path: bool = True) -> str:
    file_path = dirname(name)
    name = basename(name)
    position = name.find('?')
    if position == 0:
        name = 'image.png'  # fake image name
    elif position > 0:
        name = name[:position]
    return str(path_join(file_path, name) if save_path else name)


def is_file(_path) -> bool:
    return Path(_path).is_file()


def is_dir(_path) -> bool:
    return Path(_path).is_dir()


def basename(_path) -> str:
    return str(Path(_path).name)


def dirname(_path) -> str:
    return str(Path(_path).parent)


def path_join(_path, *args) -> str:
    return str(Path(_path).joinpath(*args))


def unlink(_path, allow_not_empty=False):
    if is_dir(_path):
        if allow_not_empty:
            rmtree(_path)
        else:
            Path(_path).rmdir()
    elif is_file(_path):
        Path(_path).unlink()


def os_stat(_path):
    if is_file(_path):
        return Path(_path).stat()
    return None


def file_size(_path):
    """
    :param _path:
    :return:
    :rtype: int
    """
    data = os_stat(_path)
    if data:
        return data.st_size
    return None


def rename(_from, _to):
    if is_file(_from) or is_dir(_from):
        is_dir(dirname(_to)) or makedirs(dirname(_to))
        move(_from, _to)


def storage(_path) -> str:
    _path = get_temp_path('storage', _path)
    make_dirs(dirname(_path))
    return str(_path)


def listing(_path) -> dict:
    """
    :param _path:
    :return: {'directories': [], 'files': []}
    """
    _dirname, _dirnames, _filenames = walk(_path)
    return {'directories': _dirnames, 'files': _filenames}


def __get_info(_path):
    try:
        with open(path_join(_path, 'info.json'), 'r') as r:
            return json_loads(r.read())
    except FileNotFoundError:
        return None


def get_info(_path) -> dict:
    """
    listing subdirectories and reading info.json files
    :param _path: [{..}, {..}, {..}]
    :return:
    """
    result = {}
    for d in listing(_path)['directories']:
        directory = path_join(_path, d)
        info = __get_info(directory)
        if info is not None:
            result[directory] = info
    return result


def __dirname(_path) -> str:
    if not is_dir(_path):
        _path = __dirname(dirname(_path))
    return str(_path)


def _disk_stat_posix(_path) -> dict:
    import os
    st = os.statvfs(_path)
    free = st.f_bavail * st.f_frsize
    total = st.f_blocks * st.f_frsize
    used = (st.f_blocks - st.f_bfree) * st.f_frsize
    return {'total': total, 'used': used, 'free': free}


def _disc_stat_win(_path) -> dict:
    import ctypes
    _, total, free = ctypes.c_ulonglong(), ctypes.c_ulonglong(), ctypes.c_ulonglong()
    fun = ctypes.windll.kernel32.GetDiskFreeSpaceExA
    ret = fun(_path, ctypes.byref(_), ctypes.byref(total), ctypes.byref(free))
    if ret == 0:
        fun = ctypes.windll.kernel32.GetDiskFreeSpaceExW
        ret = fun(_path, ctypes.byref(_), ctypes.byref(total), ctypes.byref(free))
        if ret == 0:
            raise ctypes.WinError()
    used = total.value - free.value
    return {'total': total.value, 'used': used, 'free': free.value}


def get_disk_stat(_path) -> dict:
    import os
    _path = __dirname(_path)

    if hasattr(os, 'statvfs'):  # POSIX
        return _disk_stat_posix(_path)
    elif os.name == 'nt':  # Windows
        return _disc_stat_win(_path)
    else:
        raise NotImplementedError('Platform not supported')


def check_free_space(_path: str, min_size: int = 100, percent: bool = False) -> bool:
    """
    min_size = 10  # percent = True
    min_size = 10  # percent = False (default)

    :param _path:
    :param min_size:
    :param percent:
    :return:
    """
    _stat = get_disk_stat(_path)
    if percent:
        _free = _stat['free'] / _stat['total']
        if (_free * 100) < min_size:
            return False
        return True
    else:
        _free = _stat['free'] / (2 << 19)  # 1Mb
        if _free < min_size:
            return False
        return True




from ._http2 import Http2


class Std:
    def _elements(self, selector, content=None) -> list:
        if not content:
            content = self.content
        return self.document_fromstring(content, selector)

    def _cover_from_content(self, selector, attr='src') -> str:
        image = self._elements(selector)
        if image and len(image):
            return self.http().normalize_uri(image[0].get(attr))

    @staticmethod
    def _first_select_options(parser, selector, skip_first=True) -> list:
        options = 'option'
        if skip_first:
            options = 'option + option'
        select = parser.cssselect(selector)
        if select:
            return select[0].cssselect(options)
        return []

    @classmethod
    def _images_helper(cls, parser, selector, attr='src') -> list:
        image = parser.cssselect(selector)
        return [i.get(attr).strip(' \r\n\t\0') for i in image]

    @classmethod
    def _idx_to_x2(cls, idx, default=0) -> list:
        return [
            str(idx[0]),
            str(default if len(idx) < 2 or not idx[1] else idx[1])
        ]

    @staticmethod
    def _join_groups(idx, glue='-') -> str:
        result = []
        for i in idx:
            if i:
                result.append(i)
        return glue.join(result)

    def _get_name(self, selector, url=None) -> str:
        if url is None:
            url = self.get_url()
        return self.re.search(selector, url).group(1)

    def _get_content(self, selector) -> str:
        return self.http_get(selector.format(self.domain, self.manga_name))

    def _base_cookies(self, url=None):
        if url is None:
            url = self.get_url()
        cookies = self.http().get_base_cookies(url)
        self._storage['cookies'] = cookies.get_dict()

    def parse_background(self, image) -> str:
        selector = r'background.+?url\([\'"]?([^\s]+?)[\'"]?\)'
        url = self.re.search(selector, image.get('style'))
        return self.http().normalize_uri(url.group(1))

    @property
    def content(self):
        content = self._storage.get('main_content', None)
        if content is None:
            content = self.get_main_content()
        return content

    @property
    def manga_name(self) -> str:
        name = self._storage.get('manga_name', None)
        if name is None:
            name = self.get_manga_name()
        return name

    def normal_arc_name(self, idx):
        fmt = 'vol_{:0>3}'
        if not isinstance(idx, list):
            idx = [idx]
        if len(idx) > 1:
            fmt += '-{}' * (len(idx) - 1)
        elif self._zero_fill:
            idx.append('0')
            fmt += '-{}'
        return fmt.format(*idx)

    def text_content(self, content, selector, idx: int = 0, strip: bool = True):
        doc = self.document_fromstring(content, selector)
        if not doc:
            return None
        text = doc[idx].text_content()
        if strip:
            text = text.strip()
        return text



from manga_py.crypt.base_lib import BaseLib
from manga_py.provider import Provider
from .helpers.std import Std


class WebtoonBamtokiCom(Provider, Std):

    def get_archive_name(self) -> str:
        return self.normal_arc_name(self.get_chapter_index())

    def get_chapter_index(self) -> str:
        re = self.re.compile(r'.+-(\d+).html')
        return re.search(self.chapter).group(1)

    def get_main_content(self):
        return self._get_content('{}/{}')

    def get_manga_name(self) -> str:
        url = self.get_url()
        _selector = r'\.(?:com|se)/(.+)'
        if ~url.find('.html'):
            _selector += r'-\d+\.html'
        return self._get_name(_selector)

    def get_chapters(self):
        return self._elements('.list-details a.ellipsis')

    def get_files(self):
        data = self.html_fromstring(self.chapter, '#tooncontentdata', 0)
        content = BaseLib.base64decode(data.text_content().strip('\n\t\r\0 '))
        parser = self.document_fromstring(content)
        return self._images_helper(parser, 'img')

    def get_cover(self) -> str:
        return self._cover_from_content('.title-section-inner .col-md-6 > img')

    def book_meta(self) -> dict:
        # todo meta
        pass


main = WebtoonBamtokiCom



from manga_py.provider import Provider
from .helpers.std import Std


class BatoTo(Provider, Std):

    def get_archive_name(self) -> str:
        return self.normal_arc_name([
            self.chapter_id,
            self.chapter[1]
        ])

    def get_chapter_index(self) -> str:
        return self.re.search(r'/chapter/(\d+)', self.chapter[0]).group(1)

    def get_main_content(self):
        url = self.get_url()
        if ~url.find('/chapter/'):
            url = self.html_fromstring(url, '.nav-path .nav-title > a', 0).get('href')
        return self.http_get(url)

    def get_manga_name(self) -> str:
        selector = '.nav-path .nav-title > a,.title-set .item-title > a'
        content = self.http_get(self.get_url())
        return self.text_content(content, selector, 0)

    def get_chapters(self):
        items = self._elements('.main > .item > a')
        n = self.http().normalize_uri
        return [(
            n(i.get('href')),
            i.cssselect('b')[0].text_content().strip(' \n\t\r'),
        ) for i in items]

    @staticmethod
    def _sort_files(data):
        keys = sorted(data, key=lambda _: int(_))
        return [data[i] for i in keys]

    def get_files(self):
        data = self.re.search(r'\simages\s*=\s*({.+});', self.http_get(self.chapter[0]))
        try:
            return self._sort_files(self.json.loads(data.group(1)))
        except ValueError:
            return []

    def get_cover(self) -> str:
        return self._cover_from_content('.attr-cover img')

    def book_meta(self) -> dict:
        # todo meta
        pass

    def chapter_for_json(self):
        return self.chapter[0]


main = BatoTo



from manga_py.provider import Provider


class EHentaiOrg:
    provider = None

    def __init__(self, provider: Provider):
        self.provider = provider

    def get_pages_count(self, parser):
        selector = '.gtb table.ptt td[onclick] > a'
        paginate = parser.cssselect(selector)
        max_idx = 0
        for i in paginate:
            idx = self.provider.re.search(r'\?p=(\d+)', i.get('href'))
            max_idx = max(max_idx, int(idx.group(1)))
        return max_idx

    def get_image(self, i):
        url = i.get('href')
        src = self.provider.html_fromstring(url, 'img#img', 0)
        return src.get('src')

    def get_url(self):
        url = self.provider.get_url()
        if ~url.find('?'):
            url = url[:url.find('?')]
        if ~url.find('#'):
            url = url[:url.find('#')]
        return url



from .abstract import Abstract
from .archive import Archive
from .base import Base
from .callbacks import Callbacks
from .cf_protect import CloudFlareProtect
from .chapter_helper import ChapterHelper
from .static import Static
from .web_driver import WebDriver



from manga_py.provider import Provider
from .helpers.std import Std


class MangaTubeMe(Provider, Std):
    def get_archive_name(self) -> str:
        idx = self.get_chapter_index()
        return self.normal_arc_name(idx.split('-'))

    def get_chapter_index(self) -> str:
        txt = self.chapter[0].strip()
        idx = self.re.search(r'(?:.*?)(\d+(?:\.\d+)?)', txt)
        if not idx:
            return str(self.chapter_id)
        return '-'.join(self._idx_to_x2(idx.group(1).split('.')))

    def get_main_content(self):
        return self._get_content('{}/series/{}/')

    def get_manga_name(self) -> str:
        return self._get_name('/series/([^/]+)/')

    def get_chapters(self):
        items = self._elements('#chapter li > a')
        return [(i.text_content(), i.get('href')) for i in items]

    def get_files(self):
        n = self.http().normalize_uri
        content = self.http_get(n(self.chapter[1]))
        img_path = self.re.search(r'img_path[\'"]?:\s[\'"](.+)[\'"]', content)
        img_path = n(img_path.group(1))
        images = self.re.search(r'pages[\'"]?:\s(\[\{.+\}\])', content)
        images = self.json.loads(images.group(1))
        return ['{}{}'.format(img_path, i.get('file_name')) for i in images]

    def get_cover(self):
        return self._cover_from_content('.owl-carousel img', 'data-original')

    def book_meta(self) -> dict:
        # todo meta
        pass

    def chapter_for_json(self):
        return self.chapter[1]


main = MangaTubeMe
